@@ -1,4 +1,4 @@
-# Copyright (C) 2018-'20  Frank Sachsenheim
+# Copyright (C) 2018-'21  Frank Sachsenheim
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License as published
@@ -19,9 +19,10 @@ The ``core_loaders`` module provides a set loaders to retrieve documents from va
 data sources.
 """
 
+from __future__ import annotations
 
 from copy import deepcopy
-from io import IOBase
+from io import IOBase, UnsupportedOperation
 from pathlib import Path
 from types import SimpleNamespace
 from typing import cast, Any, IO
@@ -40,11 +41,11 @@ def tag_node_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
     descendant nodes.
     """
     if isinstance(data, TagNode):
-        tree = etree.ElementTree(parser=config.parser)
+        tree = etree.ElementTree()
         root = data.clone(deep=True)
         tree._setroot(root._etree_obj)
-        utils.copy_root_siblings(data._etree_obj, root._etree_obj)
-        return tree, root._wrapper_cache
+        utils._copy_root_siblings(data._etree_obj, root._etree_obj)
+        return tree
     return "The input value is not a TagNode instance."
 
 
@@ -55,9 +56,9 @@ def etree_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
     :class:`lxml.etree._ElementTree` instances.
     """
     if isinstance(data, etree._ElementTree):
-        return deepcopy(data), {}
+        return deepcopy(data)
     if isinstance(data, etree._Element):
-        return etree.ElementTree(element=deepcopy(data), parser=config.parser), {}
+        return etree.ElementTree(element=deepcopy(data), parser=config.parser)
     return "The input value is neither an etree.Element or …Tree instance."
 
 
@@ -81,7 +82,11 @@ def buffer_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
     This loader loads a document from a :term:`file-like object`.
     """
     if isinstance(data, IOBase):
-        return etree.parse(cast(IO, data), parser=config.parser), {}
+        try:
+            data.seek(0)
+        except UnsupportedOperation:
+            pass
+        return etree.parse(cast(IO, data), parser=config.parser)
     return "The input value is no buffer object."
 
 
@@ -93,7 +98,7 @@ def ftp_http_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
     """
     if isinstance(data, str) and data.lower().startswith(("http://", "ftp://")):
         config.source_url = data
-        return etree.parse(data, parser=config.parser), {}
+        return etree.parse(data, parser=config.parser)
     return "The input value is not an URL with the ftp or http scheme."
 
 
@@ -106,7 +111,7 @@ def text_loader(data: Any, config: SimpleNamespace) -> LoaderResult:
         data = data.encode()
     if isinstance(data, bytes):
         root = etree.fromstring(data, config.parser)
-        return etree.ElementTree(element=root), {}
+        return etree.ElementTree(element=root)
     return "The input value is not a byte sequence."
 
 
