@@ -1,4 +1,4 @@
-# Copyright (C) 2018-'20  Frank Sachsenheim
+# Copyright (C) 2018-'21  Frank Sachsenheim
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License as published
@@ -15,25 +15,23 @@
 
 
 from types import SimpleNamespace
-from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Union
 
 from lxml import etree
 
 if TYPE_CHECKING:
-    from _delb.nodes import _ElementWrappingNode, NodeBase, _TagDefinition  # noqa: F401
+    from _delb.nodes import NodeBase, _TagDefinition
 
 
-ElementAttributes = etree._Attrib
 Filter = Callable[["NodeBase"], bool]
 NodeSource = Union[str, "NodeBase", "_TagDefinition"]
-_WrapperCache = Dict[int, "_ElementWrappingNode"]
 
-LoaderResult = Union[Tuple[Optional[etree._ElementTree], _WrapperCache], str]
+LoaderResult = Union[Optional[etree._ElementTree], str]
 Loader = Callable[[Any, SimpleNamespace], LoaderResult]
+LoaderConstraint = Union[Loader, Iterable[Loader], None]
 
 
 __all__ = (
-    "ElementAttributes",
     "Filter",
     "Loader",
     "LoaderResult",
