@@ -1,4 +1,4 @@
-# Copyright (C) 2018-'20  Frank Sachsenheim
+# Copyright (C) 2018-'21  Frank Sachsenheim
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License as published
@@ -18,7 +18,7 @@ from collections.abc import MutableSequence
 from copy import deepcopy
 from pathlib import Path
 from types import SimpleNamespace
-from typing import Any, Dict, Iterable, Iterator, Optional, Tuple, Union
+from typing import Any, Dict, Iterable, Iterator, Optional, Tuple, Type, Union
 from typing import IO as IOType
 
 from lxml import etree
@@ -26,12 +26,12 @@ from lxml import etree
 from _delb.exceptions import FailedDocumentLoading, InvalidOperation
 from _delb.plugins import (
     core_loaders,
-    DocumentExtensionHooks,
+    DocumentMixinHooks,
     plugin_manager as _plugin_manager,
 )
 from _delb.nodes import (
-    _get_or_create_element_wrapper,
     _is_tag_or_text_node,
+    _wrapper_cache,
     altered_default_filters,
     any_of,
     is_comment_node,
@@ -51,9 +51,11 @@ from _delb.nodes import (
     TagNode,
     TextNode,
 )
-from _delb.typing import Loader, _WrapperCache
+from _delb.typing import Loader
 from _delb.utils import (
-    copy_root_siblings,
+    DEFAULT_PARSER,
+    _collect_subclasses,
+    _copy_root_siblings,
     first,
     get_traverser,
     last,
@@ -67,20 +69,12 @@ from _delb.utils import (
 _plugin_manager.load_plugins()
 
 
-# constants
-
-
-# care has to be taken:
-# https://wiki.tei-c.org/index.php/XML_Whitespace#Recommendations
-# https://wiki.tei-c.org/index.php/XML_Whitespace#Default_Whitespace_Processing
-# https://lxml.de/FAQ.html#why-doesn-t-the-pretty-print-option-reformat-my-xml-output
-DEFAULT_PARSER = etree.XMLParser(remove_blank_text=True)
-
-
 # api
 
 
 class _RootSiblingsContainer(ABC, MutableSequence):
+    __slots__ = ("_document",)
+
     def __init__(self, document: "Document"):
         self._document = document
 
@@ -184,7 +178,7 @@ class DocumentMeta(type):
                     (f"{x[0]}:\n\n{x[1]}" for x in extension_docs)
                 )
 
-        base_classes += extension_classes + (DocumentExtensionHooks,)
+        base_classes += extension_classes + (DocumentMixinHooks,)
 
         namespace["_loaders"] = (
             core_loaders.tag_node_loader,
@@ -220,59 +214,71 @@ class Document(metaclass=DocumentMeta):
 
     :param source: Anything that the configured loaders can make sense of to return a
                    parsed document tree.
+    :param collapse_whitespace: :meth:`Collapses the content's whitespace
+                                <delb.Document.collapse_whitespace>` after loading the
+                                document.
     :param parser: An optional :class:`lxml.etree.XMLParser` instance that is used to
                    parse a document stream.
+    :param klass: Explicitly define the initilized class. This can be useful for
+                  applications that have :ref:`default document subclasses
+                  <extending-subclasses>` in use.
     :param config: Additional keyword arguments for the configuration of extension
                    classes.
     """
 
     _loaders: Tuple[Loader, ...]
-    __slots__ = ("__root_node__", "head_nodes", "source_url", "tail_nodes")
+    __slots__ = ("__root_node__", "config", "head_nodes", "source_url", "tail_nodes")
 
-    def __init__(self, source: Any, parser: etree.XMLParser = DEFAULT_PARSER, **config):
-        self.config: SimpleNamespace = SimpleNamespace()
-        """
-        Beside the used ``parser``, this property contains the namespaced data that
-        extension classes and loaders may have stored.
-        """
-        self.config.parser = parser
-        self._init_config(config)  # type: ignore
-        if config:
-            raise RuntimeError(
-                "Not all configuration arguments have been processed. You either "
-                "passed invalid arguments or an extension doesn't handle them "
-                f"properly: {config}"
-            )
+    def __new__(
+        cls,
+        source,
+        collapse_whitespace=False,
+        parser=DEFAULT_PARSER,
+        klass=None,
+        **config,
+    ):
+        config = cls.__process_config(collapse_whitespace, parser, config)
+        root = cls.__load_source(source, config)
 
-        loader_excuses: Dict[Loader, Union[str, Exception]] = {}
-        loaded_tree: Optional[etree._ElementTree] = None
-        wrapper_cache: Optional[_WrapperCache] = None
-        for loader in self._loaders:
-            try:
-                loader_result = loader(source, self.config)
-            except Exception as e:
-                loader_excuses[loader] = e
-            else:
-                if isinstance(loader_result, tuple):
-                    loaded_tree, wrapper_cache = loader_result
+        if klass is None:
+            subclasses = list()
+            _collect_subclasses(cls, subclasses)
+
+            for subclass in subclasses:
+                if hasattr(subclass, "__class_test__") and subclass.__class_test__(
+                    root, config
+                ):
+                    klass = subclass
                     break
-                else:
-                    loader_excuses[loader] = loader_result
+            else:
+                klass = cls
 
-        if loaded_tree is None:
-            raise FailedDocumentLoading(source, loader_excuses)
+        assert issubclass(klass, Document)
+        instance = super().__new__(klass)
+        instance.config = config
+        instance.root = root
+        return instance
 
-        assert isinstance(loaded_tree, etree._ElementTree)
-        assert isinstance(wrapper_cache, dict)
+    def __init__(
+        self,
+        source: Any,
+        collapse_whitespace: bool = False,
+        parser: etree.XMLParser = DEFAULT_PARSER,
+        klass: Optional[Type["Document"]] = None,
+        **config,
+    ):
+        self.config: SimpleNamespace
+        """
+        Beside the used ``parser`` and ``collapsed_whitespace`` option, this property
+        contains the namespaced data that extension classes and loaders may have stored.
+        """
+        if collapse_whitespace:
+            self.collapse_whitespace()
 
-        root = _get_or_create_element_wrapper(loaded_tree.getroot(), wrapper_cache)
-        assert isinstance(root, TagNode)
-        self.root = root
         self.source_url: Optional[str] = self.config.__dict__.pop("source_url", None)
         """
         The source URL where a loader obtained the document's contents or ``None``.
         """
-
         self.head_nodes = _HeadNodes(self)
         """
         A list-like accessor to the nodes that precede the document's root node.
@@ -284,14 +290,46 @@ class Document(metaclass=DocumentMeta):
         Note that nodes can't be removed or replaced.
         """
 
+    @classmethod
+    def __process_config(cls, collapse_whitespace, parser, kwargs) -> SimpleNamespace:
+        config = SimpleNamespace(collapse_whitespace=collapse_whitespace, parser=parser)
+        cls._init_config(config, kwargs)  # type: ignore
+        return config
+
+    @classmethod
+    def __load_source(cls, source: Any, config: SimpleNamespace) -> TagNode:
+        loader_excuses: Dict[Loader, Union[str, Exception]] = {}
+
+        for loader in cls._loaders:
+            try:
+                loader_result = loader(source, config)
+            except Exception as e:
+                loader_excuses[loader] = e
+            else:
+                if isinstance(loader_result, etree._ElementTree):
+                    loaded_tree = loader_result
+                    break
+                else:
+                    assert isinstance(loader_result, str)
+                    loader_excuses[loader] = loader_result
+        else:
+            raise FailedDocumentLoading(source, loader_excuses)
+
+        assert isinstance(loaded_tree, etree._ElementTree)
+
+        root = _wrapper_cache(loaded_tree.getroot())
+        assert isinstance(root, TagNode)
+
+        return root
+
     def __contains__(self, node: NodeBase) -> bool:
         return node.document is self
 
     def __str__(self):
         cloned_root = self.root.clone(deep=True)
         cloned_root.merge_text_nodes()
-        copy_root_siblings(self.root._etree_obj, cloned_root._etree_obj)
-        return etree.tounicode(cloned_root._etree_obj.getroottree())
+        _copy_root_siblings(self.root._etree_obj, cloned_root._etree_obj)
+        return etree.tostring(cloned_root._etree_obj.getroottree(), encoding="unicode")
 
     def cleanup_namespaces(
         self,
@@ -333,13 +371,24 @@ class Document(metaclass=DocumentMeta):
         """
         :return: Another instance with the duplicated contents.
         """
+        result = Document(self.root, klass=self.__class__)
         # lxml.etree.XMLParser instances aren't pickable / copyable
         parser = self.config.__dict__.pop("parser")
-        result = self.__class__(self.root, parser=parser)
         result.config = deepcopy(self.config)
         self.config.parser = result.config.parser = parser
         return result
 
+    def collapse_whitespace(self):
+        """
+        Collapses whitespace as described here:
+        https://wiki.tei-c.org/index.php/XML_Whitespace#Recommendations
+
+        Implicitly merges all neighbouring text nodes.
+        """
+        self.merge_text_nodes()
+        with altered_default_filters():
+            self.root._collapse_whitespace()
+
     def css_select(self, expression: str) -> QueryResults:
         """
         This method proxies to the :meth:`TagNode.css_select` method of the document's
@@ -355,7 +404,7 @@ class Document(metaclass=DocumentMeta):
         self.root.merge_text_nodes()
 
     @property
-    def namespaces(self) -> Dict[str, str]:
+    def namespaces(self) -> Dict[Optional[str], str]:
         """
         The namespace mapping of the document's :attr:`root <Document.root>` node.
         """
@@ -377,8 +426,8 @@ class Document(metaclass=DocumentMeta):
 
     @property
     def root(self) -> "TagNode":
-        """ The root node of a document tree. """
-        return getattr(self, "__root_node__")
+        """The root node of a document tree."""
+        return self.__root_node__
 
     @root.setter
     def root(self, node: "TagNode"):
@@ -397,11 +446,11 @@ class Document(metaclass=DocumentMeta):
 
         current_root = getattr(self, "__root_node__", None)
         if current_root is not None:
-            copy_root_siblings(current_root._etree_obj, node._etree_obj)
-            delattr(current_root, "__document__")
+            _copy_root_siblings(current_root._etree_obj, node._etree_obj)
+            current_root.__document__ = None
 
-        setattr(node, "__document__", self)
-        setattr(self, "__root_node__", node)
+        self.__root_node__ = node
+        node.__document__ = self
 
     def save(self, path: Path, pretty: bool = False, **cleanup_namespaces_args):
         """
